/**
 * Copyright (C) 1998-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * <p>
 * A "Just-in-time" iterator for streaming a flow of objects across an iterable interface.
 * This class is useful in situations where you don't want to keep your list of objects around
 * in memory, but instead are using the collection to move them from one location (often slow-loading)
 * to another.
 * </p>
 * <p>
 * To use the jiterator, just construct it and starting pushing items into it. Once the last item has
 * been added, call {@link #complete()}.
 * </p>
 * <p>
 * Last Modified $Date: 2009/07/02 01:37:02 $
 * </p>
 * @author George Reese
 * @param <T> the type of object being managed in the jiterator
 */
public class Jiterator<T> implements Iterator<T>, Iterable<T> {
    private JiteratorFilter<T> filter        = null;
    private long               lastTouch     = System.currentTimeMillis();
    private Exception          loadException = null;
    private boolean            loaded        = false;
    private ArrayList<T>       waiting       = new ArrayList<T>();
    
    private transient boolean nexting = false;
    
    /**
     * Constructs an empty jiterator into which you can start adding items.
     */
    public Jiterator() {
        // no-op
    }

    public Jiterator(JiteratorFilter<T> filter) {
        this.filter = filter;
    }
    
    /**
     * Constructs a jiterator from the specified collection. This jiterator will load
     * items from the source collection in a background thread. It will <b>not</b>
     * make the jiterator complete. You can thus add elements in addition to the source
     * collection.
     * @param list the source list of items to initialize the jiterator with
     */
    public Jiterator(Collection<T> list) {
        final Collection<T> flist = list;
        
        Thread t = new Thread() {
            public void run() {
                for( T item : flist ) {
                    push(item);
                }
            }
        };
        t.setName("Jiterator loader");
        t.setDaemon(true);
        t.start();
    }
    
    /**
     * Marks the jiterator as complete. Once you call this method, you cannot add any more
     * items into the jiterator. If you fail to call this method, any threads reading from this
     * jiterator will ultimately hang until you call this method.
     */
    public synchronized void complete() {
        if( loaded ) {
            return;
        }
        loaded = true;
        lastTouch = System.currentTimeMillis();
        if( waiting != null && waiting.isEmpty() ) {
            waiting = null;
        }
        notifyAll();
    }
    
    /**
     * Checks to see if there are more elements to be processed in the jiterator. If this method
     * is called prior to the jiterator being loaded with an item, it will hang until
     * either an item is added into the jiterator or the jiterator is marked as complete.
     * @return true if there are more elements to be processed
     * @throws JiteratorLoadException an error occurred during the load of the jiterator
     */
    public synchronized boolean hasNext() {
        while( nexting ) {
            try { wait(150L); }
            catch( InterruptedException e ) { /* ignore */ }
        }
        if( loadException != null ) {
            throw new JiteratorLoadException(loadException);
        }
        nexting = true;
        try {
            if( waiting == null ) {
                if( loadException != null ) {
                    throw new JiteratorLoadException(loadException);                    
                }
                return false;
            }
            if( !waiting.isEmpty() ) {
                return true;
            }
            waitForPush();
            if( waiting == null ) {
                if( loadException != null ) {
                    throw new JiteratorLoadException(loadException);                    
                }                
                return false;
            }
            return !waiting.isEmpty();
        }
        finally {
            nexting = false;
            notifyAll();
        }
    }

    /** 
     * @return true if all items have been processed and this jiterator can be discarded
     */
    public synchronized boolean isComplete() {
        return (waiting == null);
    }
    
    /**
     * @return true if all items have been loaded into the jiterator
     */
    public synchronized boolean isLoaded() {
        return loaded;
    }

    /**
     * @return this jiterator
     */
    public Iterator<T> iterator() {
        if( loadException != null ) {
            throw new JiteratorLoadException(loadException);
        }
        return this;
    }
    
    /**
     * Provides the next element in the jiterator. If this method
     * is called prior to the jiterator being loaded with an item, it will hang until
     * either an item is added into the jiterator or the jiterator is marked as complete.
     * @return true if there are more elements to be processed
     * @throws NoSuchElementException an attempt was made to read beyond the last item in the jiterator
     */    
    public synchronized T next() throws JiteratorLoadException {
        while( nexting ) {
            try { wait(150L); }
            catch( InterruptedException e ) { /* ignore */ }
        }
        nexting = true;
        try {
            T t;
            
            if( waiting == null ) {
                if( loadException != null ) {
                    throw new JiteratorLoadException(loadException);
                }
                throw new NoSuchElementException("Invalid attempt to get another element from empty iterator.");
            }
            if( !waiting.isEmpty() ) {
                if( loadException != null ) {
                    throw new JiteratorLoadException(loadException);
                }
                t = waiting.get(0);
                waiting.remove(0);
                if( waiting.isEmpty() && loaded ) { 
                    waiting = null;
                } 
                return t;
            }
            waitForPush();
            if( waiting == null || waiting.isEmpty() ) {
                throw new NoSuchElementException("Invalid attempt to get another element from empty iterator.");            
            }
            if( loadException != null ) {
                throw new JiteratorLoadException(loadException);
            }
            t = waiting.get(0);
            waiting.remove(0);
            if( waiting.isEmpty() && loaded ) { 
                waiting = null;
            }
            return t;        
        }
        finally {
            nexting = false;
            notifyAll();
        }
    }

    /**
     * Pushes a new item into the jiterator. The new item will be added to the end of the jiterator.
     * @param item the item to be added
     * @throws IllegalStateException an attempt was made to push an item on a jiterator marked complete
     */
    public synchronized void push(T item) {
        if( waiting == null ) {
            throw new IllegalStateException("Invalid attempt to add an item to a completed list.");
        }
        if( filter != null ) {
            try {
                if( !filter.filter(item) ) {
                    return;
                }
            }
            catch( Throwable t ) {
                Exception e;
                
                if( t instanceof Exception ) {
                    e = (Exception)t;
                }
                else {
                    e = new RuntimeException(t);
                }
                setLoadException(e);
                return;
            }
        }
        waiting.add(item);
        lastTouch = System.currentTimeMillis();
        notifyAll();
    }
    
    /**
     * This operation is not allowed
     * @throws UnsupportedOperationException always thrown
     */
    public synchronized void remove() {
        throw new UnsupportedOperationException("Removing from a jiterator is not supported.");
    }

    public synchronized void setLoadException(Exception e) {
        loadException  = e;
        loaded = true;
        waiting = null;
        nexting = false;
        lastTouch = System.currentTimeMillis();
        notifyAll();
    }
    
    private transient long scream = 0L;
    /**
     * Waits for a new item to be put into the jiterator or for the jiterator to be marked empty.
     */
    private synchronized void waitForPush() {
        long waitStart = -1L;
        
        while( waiting != null && waiting.isEmpty() && !loaded ) {
            if( waitStart == -1L ) {
                waitStart = System.currentTimeMillis();
            }
            else {
                if( (System.currentTimeMillis() - waitStart) > CalendarWrapper.MINUTE ) {
                    if( (System.currentTimeMillis() - scream) > CalendarWrapper.MINUTE ) {
                        scream = System.currentTimeMillis();
                        System.out.println(this + ": " + ((System.currentTimeMillis()-lastTouch)/1000) + " seconds since last touch.");
                        System.out.println("\tLoad exception: " + loadException);
                        System.out.println("\tLoaded: " + loaded);
                        System.out.println("\tWaiting: " + waiting);
                    }
                }
            }
            try { wait(150L); }
            catch( InterruptedException e ) { /* ignore */ }
        }        
    }
}
